import CodeMirror, { wheelEventPixels } from "codemirror";
const CLASS_NAME = "cmw-scrollbar";
// - 3*i = lane
// - 3*i + 1 = start line
// - 3*i + 2 = end line
// interface OverviewDecoration { [color: string]: number[]; }
const MIN_DECORATION_HEIGHT = 6;
// Based on `addon/scroll/simplescrollbars`.
// Simulates VSCode's "Overview Ruler".
// Marks go into 3 columns.
// 1. Content Changes (added, removed, modified)
// 2. Document Highlights (word, range, selection)
// 3. Diagnostics (error, warning, info)
class ScrollbarsWithOverview {
    constructor(place, scroll, cm) {
        this.addClass = CLASS_NAME;
        this.horiz = new Bar("horizontal", scroll);
        this.horiz.node.style.cssText = "position: absolute; z-index: 6;";
        place(this.horiz.node);
        this.vert = new Bar("vertical", scroll);
        this.vert.node.style.cssText = "position: absolute; z-index: 6;";
        place(this.vert.node);
        this.width = null;
        this.scrollHeight = 0;
        this.viewHeight = 0;
        // TODO Make colors customizable with CSS custom props.
        // `getComputedStyle(document.documentElement).getPropertyValue('--cmw-color-foo')`;
        const wrapper = cm.getWrapperElement();
        const cursorCanvas = document.createElement("canvas");
        this.cursorCanvas = cursorCanvas;
        cursorCanvas.style.cssText = [
            "position: absolute;",
            "transfrom: translate3d(0, 0, 0);",
            "right: 0;",
            "top: 0;",
            // Must be lower than the scrollbars to make click/drag work.
            "z-index: 5;",
        ].join("");
        cursorCanvas.width = 12;
        cursorCanvas.height = wrapper.clientHeight;
        wrapper.appendChild(cursorCanvas);
        const cursorCtx = cursorCanvas.getContext("2d");
        cursorCtx.fillStyle = "rgba(73, 80, 87, 0.7)";
        const lineOffset = (cm, n) => {
            const prev = n > 0 ? cm.defaultTextHeight() * n : 0;
        };
        const drawCursorLine = (cm) => {
            const n = cm.lineCount();
            if (n === 0)
                return;
            if (this.scrollHeight === 0) {
                cursorCtx.clearRect(0, 0, cursorCanvas.width, cursorCanvas.height);
                cursorCtx.fillRect(0, 0, cursorCanvas.width, 2);
                return;
            }
            const heightRatio = cursorCanvas.height / this.scrollHeight;
            console.log("scrollHeight", this.scrollHeight);
            console.log("viewHeight", this.viewHeight);
            console.log("heightRatio", heightRatio);
            console.log("height at line", cm.heightAtLine(cm.getCursor().line, "local"));
            console.log("height at line", cm.heightAtLine(cm.lastLine(), "local"));
            const dy = (cursorCanvas.height / n) * cm.getCursor().line;
            cursorCtx.clearRect(0, 0, cursorCanvas.width, cursorCanvas.height);
            cursorCtx.fillRect(0, dy, cursorCanvas.width, 2);
        };
        drawCursorLine(cm);
        cm.on("cursorActivity", drawCursorLine);
        // TODO
        // cm.on("cmw:beforeDiagnostics", clearDiagnosticsCanvas);
        // cm.on("cmw:diagnostic", markDiagnosticCanvas);
        // cm.on("cmw:removeDiagnostics", clearDiagnosticsCanvas);
        //
        // cm.on("cmw:beforeDocumentHighlight", clearDocHighlightCanvas);
        // cm.on("cmw:documentHighlight", markHighlight);
        // cm.on("cmw:removeDocumentHightlight", clearDocHighlightCanvas);
    }
    update(measure) {
        this.scrollHeight = measure.scrollHeight;
        this.viewHeight = measure.viewHeight;
        if (this.width == null) {
            const style = window.getComputedStyle(this.horiz.node);
            if (style)
                this.width = parseInt(style.height);
        }
        const width = this.width || 0;
        const needsH = measure.scrollWidth > measure.clientWidth + 1;
        // Always show the vertical scroll bar because we'll show marks on it.
        const needsV = true;
        this.vert.node.style.display = "block";
        this.horiz.node.style.display = needsH ? "block" : "none";
        // Vertical scrollbar takes full height even when horizontal scrollbar is shown
        if (needsV) {
            this.vert.update(measure.scrollHeight, measure.clientHeight, measure.viewHeight);
            this.vert.node.style.bottom = "0";
        }
        if (needsH) {
            this.horiz.update(measure.scrollWidth, measure.clientWidth, measure.viewWidth - width - measure.barLeft);
            this.horiz.node.style.right = width + "px";
            this.horiz.node.style.left = measure.barLeft + "px";
        }
        return { right: width, bottom: 0 };
    }
    setScrollTop(pos) {
        this.vert.setPos(pos);
    }
    setScrollLeft(pos) {
        this.horiz.setPos(pos);
    }
    clear() {
        this.horiz.node.remove();
        this.vert.node.remove();
        this.cursorCanvas.remove();
    }
}
// Expose this scrollbar for `scrollbarStyle`
CodeMirror.scrollbarModel.cmw = ScrollbarsWithOverview;
class Bar {
    constructor(orientation, scroll) {
        this.orientation = orientation;
        this.scroll = scroll;
        this.screen = 1;
        this.total = 1;
        this.size = 1;
        this.pos = 0;
        this.node = document.createElement("div");
        this.node.className = CLASS_NAME + "-" + orientation;
        this.inner = this.node.appendChild(document.createElement("div"));
        // Support dragging
        this.inner.addEventListener("mousedown", (e) => {
            if (e.button !== 0)
                return;
            e.preventDefault();
            const axis = this.orientation == "horizontal" ? "pageX" : "pageY";
            const start = e[axis];
            const startpos = this.pos;
            const done = () => {
                document.removeEventListener("mousemove", move);
                document.removeEventListener("mouseup", done);
            };
            const move = (e) => {
                if (e.button !== 0)
                    return done();
                this.moveTo(startpos + (e[axis] - start) * (this.total / this.size));
            };
            document.addEventListener("mousemove", move);
            document.addEventListener("mouseup", done);
        });
        // Move up/down by screen when clicked
        // TODO Support jumping to the position clicked
        this.node.addEventListener("click", (e) => {
            e.preventDefault();
            const innerBox = this.inner.getBoundingClientRect();
            let where;
            if (this.orientation === "horizontal") {
                where =
                    e.clientX < innerBox.left ? -1 : e.clientX > innerBox.right ? 1 : 0;
            }
            else {
                where =
                    e.clientY < innerBox.top ? -1 : e.clientY > innerBox.bottom ? 1 : 0;
            }
            this.moveTo(this.pos + where * this.screen);
        });
        this.node.addEventListener("wheel", (e) => {
            const moved = wheelEventPixels(e)[this.orientation == "horizontal" ? "x" : "y"];
            const oldPos = this.pos;
            this.moveTo(this.pos + moved);
            if (this.pos !== oldPos)
                e.preventDefault();
        });
    }
    setPos(pos, force) {
        if (pos < 0)
            pos = 0;
        if (pos > this.total - this.screen)
            pos = this.total - this.screen;
        if (!force && pos == this.pos)
            return false;
        this.pos = pos;
        this.inner.style[this.orientation == "horizontal" ? "left" : "top"] =
            pos * (this.size / this.total) + "px";
        return true;
    }
    update(scrollSize, clientSize, barSize) {
        const sizeChanged = this.screen != clientSize ||
            this.total != scrollSize ||
            this.size != barSize;
        if (sizeChanged) {
            this.screen = clientSize;
            this.total = scrollSize;
            this.size = barSize;
        }
        const minButtonSize = 10;
        let buttonSize = this.screen * (this.size / this.total);
        if (buttonSize < minButtonSize) {
            this.size -= minButtonSize - buttonSize;
            buttonSize = minButtonSize;
        }
        this.inner.style[this.orientation == "horizontal" ? "width" : "height"] =
            buttonSize + "px";
        this.setPos(this.pos, sizeChanged);
    }
    moveTo(pos) {
        if (this.setPos(pos))
            this.scroll(pos, this.orientation);
    }
}
//# sourceMappingURL=scrollbars.js.map